from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise
from slapos.util import str2bytes
from zope.interface import implementer
import datetime
import base64
import re
import ssl

from pyasn1.codec.der import decoder
from pyasn1_modules import rfc2459


def load_pem_x509_certificate(pem_data):
  """Extracts and decodes the DER bytes from PEM certificate data."""
  pem_body = re.search(
    "-----BEGIN CERTIFICATE-----"
    "(.*?)"
    "-----END CERTIFICATE-----",
    pem_data,
    re.DOTALL,
  )
  if not pem_body:
    raise ValueError("Invalid PEM encoding")
  der_bytes = base64.b64decode(
    "".join(pem_body.group(1).strip().splitlines()))
  return der_bytes


def get_notAfter_from_certificate(der_bytes):
  """Parses DER certificate and extracts the notAfter field."""
  cert, _ = decoder.decode(der_bytes, asn1Spec=rfc2459.Certificate())
  validity = cert['tbsCertificate']['validity']
  not_after_str = str(validity['notAfter'].getComponent())
  dt = datetime.datetime.strptime(not_after_str, '%y%m%d%H%M%SZ')
  return dt


@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  def sense(self):
    """
      Check the certificate
    """

    certificate_file = self.getConfig('certificate')
    key_file = self.getConfig('key', None)

    try:
      certificate_expiration_days = int(
        self.getConfig('certificate-expiration-days', '15'))
    except ValueError:
      self.logger.error('ERROR certificate-expiration-days is wrong: %r' % (
        self.getConfig('certificate-expiration-days')))
      return

    try:
      with open(certificate_file, 'r') as fh:
        pem_content = fh.read()
      der_bytes = load_pem_x509_certificate(pem_content)
      not_after = get_notAfter_from_certificate(der_bytes)
    except Exception as e:
      self.logger.error(
        'ERROR Problem loading certificate %r, error: %s' % (
          certificate_file, e))
      return

    if not_after - datetime.timedelta(
       days=certificate_expiration_days) < datetime.datetime.utcnow():
      self.logger.error(
       'ERROR Certificate %r will expire in less than %s days' % (
         certificate_file, certificate_expiration_days))
      return

    if key_file is not None:
      try:
        ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_ctx.load_cert_chain(certfile=certificate_file, keyfile=key_file)
      except Exception as e:
        if isinstance(e, ssl.SSLError) and e.reason == 'KEY_VALUES_MISMATCH':
          self.logger.error(
            'ERROR Certificate %r does not match key %r' % (
              certificate_file, key_file))
        else:
          self.logger.error(
            'ERROR Problem loading key %r, error: %s' % (key_file, e))
        return

    if key_file:
      self.logger.info(
        'OK Certificate %r and key %r are ok' % (certificate_file, key_file))
    else:
      self.logger.info(
        'OK Certificate %r is ok, no key provided' % (certificate_file,))



import mock
import os
import time
from datetime import datetime
from datetime import timedelta
from slapos.grid.promise import PromiseError
from slapos.promise.plugin.check_oru_stats_log import RunPromise
from . import TestPromisePluginMixin


class TestCheckOruConfigLog(TestPromisePluginMixin):

    promise_name = "check-oru-stats-log.py"

    def setUp(self):
        super(TestCheckOruConfigLog, self).setUp()

    def writePromise(self, **kw):
        super(TestCheckOruConfigLog, self).writePromise(self.promise_name,
            "from %s import %s\nextra_config_dict = %r\n"
            % (RunPromise.__module__, RunPromise.__name__, kw))

    def test_promise_success(self):
        self.stats_log = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'stats.log')
        if os.path.exists(self.stats_log):
            os.remove(self.stats_log)
        with open(self.stats_log, 'w+') as f:
            f.write("""2023-05-23 04:32:46,350 [INFO] Connecting to ('2a11:9ac1:6:800a::1', 830), user oranuser...
2023-05-23 04:32:48,830 [INFO] Connection to ('2a11:9ac1:6:800a::1', 830) successful
2023-05-23 04:32:49,692 [INFO] Subscription to ('2a11:9ac1:6:800a::1', 830) successful
2023-05-23 04:32:49,692 [DEBUG] Waiting for notification from ('2a11:9ac1:6:800a::1', 830)...
2023-05-23 04:33:09,787 [DEBUG] Got new notification from ('2a11:9ac1:6:800a::1', 830)...
2023-05-23 04:33:09,788 [DEBUG] Waiting for notification from ('2a11:9ac1:6:800a::1', 830)...
2023-05-23 04:33:41,318 [DEBUG] Got new notification from ('2a11:9ac1:6:800a::1', 830)...
2023-05-23 04:33:41,319 [DEBUG] Waiting for notification from ('2a11:9ac1:6:800a::1', 830)...
"""
            )
        self.writePromise(**{
            'stats-log': self.stats_log,
        })
        self.configureLauncher()
        self.launcher.run()

    def test_promise_fail(self):
        self.stats_log = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'stats.log')
        if os.path.exists(self.stats_log):
            os.remove(self.stats_log)
        with open(self.stats_log, 'w') as f:
            f.write("""
2023-05-23 04:32:33,230 [INFO] Connecting to ('2a11:9ac1:6:800a::1', 830), user oranuser...
2023-05-23 04:32:36,339 [DEBUG] Got exception, waiting 10 seconds before reconnecting...
2023-05-23 04:32:36,340 [DEBUG] Could not open socket to 2a11:9ac1:6:800a::1:830
"""
            )
        self.writePromise(**{
            'stats-log': self.stats_log,
        })
        self.configureLauncher()
        with self.assertRaises(PromiseError):
            self.launcher.run()


if __name__ == '__main__':
    unittest.main()


import itertools
import json
import logging
import os
import textwrap

from dateutil import parser as dateparser
from datetime import datetime
from slapos.grid.promise.generic import GenericPromise


def iter_reverse_lines(f):
  """
    Read lines from the end of the file
  """
  f.seek(0, os.SEEK_END)
  while True:
    try:
      while f.seek(-2, os.SEEK_CUR) and f.read(1) != b'\n':
        pass
    except OSError:
      return
    pos = f.tell()
    yield f.readline()
    f.seek(pos, os.SEEK_SET)


def iter_logrotate_file_handle(path, mode='r'):
  """
    Yield successive file handles for rotated logs
    (XX.log, XX.log.1, XX.log.2, ...)
  """
  for i in itertools.count():
    path_i = path + str(i or '')
    try:
      with open(path_i, mode) as f:
        yield f
    except OSError:
      break

def get_json_log_data_interval(json_log_file, interval):
  """
    Get all data in the last "interval" seconds from JSON log
    Reads rotated logs too (XX.log, XX.log.1, XX.log.2, ...)
  """
  current_time = datetime.now()
  data_list = []
  for f in iter_logrotate_file_handle(json_log_file, 'rb'):
    for line in iter_reverse_lines(f):
      l = json.loads(line)
      timestamp = dateparser.parse(l['time'])
      if (current_time - timestamp).total_seconds() > interval:
        return data_list
      data_list.append(l['data'])
  return data_list

def get_json_log_latest_timestamp(json_log_file):
  """
    Get latest timestamp from JSON log
    Reads rotated logs too (XX.log, XX.log.1, XX.log.2, ...)
  """
  for f in iter_logrotate_file_handle(json_log_file, 'rb'):
    for line in iter_reverse_lines(f):
      l = json.loads(line)
      return dateparser.parse(l['time']).timestamp()
  return 0


class JSONPromise(GenericPromise):
  def __init__(self, config):
    self.__name = config.get('name', None)
    self.__log_folder = config.get('log-folder', None)

    super(JSONPromise, self).__init__(config)
    json_log_name = os.path.splitext(self.__name)[0] + '.json.log'
    self.__json_log_file = os.path.join(self.__log_folder, json_log_name)
    self.json_logger = self.__make_json_logger(self.__json_log_file)

  def __make_json_logger(self, json_log_file):
    logger = logging.getLogger('json-logger')
    logger.setLevel(logging.INFO)
    handler = logging.FileHandler(json_log_file)
    formatter = logging.Formatter(
      '{"time": "%(asctime)s", "log_level": "%(levelname)s"'
      ', "message": "%(message)s", "data": %(data)s}'
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger

  def get_json_log_data_interval(self, interval):
    return get_json_log_data_interval(self.__json_log_file, interval)

def tail_file(file_path, line_count=10):
  """
  Returns the last lines of file.
  """
  line_list = []
  with open(file_path, 'rb') as f:
    BUFSIZ = 1024
    f.seek(0, 2)
    bytes = f.tell()
    size = line_count + 1
    block = -1
    while size > 0 and bytes > 0:
      if bytes - BUFSIZ > 0:
          # Seek back one whole BUFSIZ
          f.seek(block * BUFSIZ, 2)
          line_list.insert(0, f.read(BUFSIZ).decode())
      else:
          f.seek(0, 0)
          # only read what was not read
          line_list.insert(0, f.read(bytes).decode())
      line_len = line_list[0].count('\n')
      size -= line_len
      bytes -= BUFSIZ
      block -= 1
  return '\n'.join(''.join(line_list).splitlines()[-line_count:])



import json
import os
import psutil
import time

from .util import JSONPromise

from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):

  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    self.setPeriodicity(float(self.getConfig('frequency', 2)))
    self.avg_flag_file = self.getConfig('avg-flag-file', 'last_avg')
    self.max_spot_temp = float(self.getConfig('max-spot-temp', 90)) # °C
    self.max_avg_temp = float(self.getConfig('max-avg-temp', 80)) # °C
    self.avg_temp_duration = int(self.getConfig('avg-temp-duration', 600)) # secondes

  def sense(self):
    success = True
    # Get current temperature
    try:
      cpu_temp = psutil.sensors_temperatures()['coretemp'][0][1]
    except (KeyError, IndexError) as e:
      # Put logger.info to avoid errors when sensors are not 
      # supported by OS (ex: VM)
      self.logger.info("Could not read core temperature on VM")
      return

    # Check spot temperature
    if cpu_temp > self.max_spot_temp:
      success = False
      self.logger.error(
        "Temperature reached critical threshold: %s °C"
        " (threshold is %s °C)",
        cpu_temp, self.max_spot_temp)

    # Log temperature
    data = json.dumps({'cpu_temperature': cpu_temp})
    self.json_logger.info("Temperature data", extra={'data': data})

    # TODO: promise should compute average only with logs between interval
    # Computer average temperature
    avg_computation_period = self.avg_temp_duration / 4
    try:
      t = os.path.getmtime(self.avg_flag_file)
    except OSError:
      t = 0
    if (time.time() - t) > avg_computation_period:
      open(self.avg_flag_file, 'w').close()
      temp_list = self.get_json_log_data_interval(self.avg_temp_duration)
      if temp_list:
        avg_temp = sum(x['cpu_temperature'] for x in temp_list) / len(temp_list)
        if avg_temp > self.max_avg_temp:
          success = False
          self.logger.error(
            "Average temperature over the last %ds reached threshold: %s °C"
            " (threshold is %s °C)",
            self.avg_temp_duration, avg_temp, self.max_avg_temp)
      else:
        success = False
        self.logger.error("Couldn't read temperature from log")

    if success:
      self.logger.info("Temperature OK (%s °C)", cpu_temp)

  def test(self):
    """
    Called after sense() if the instance is still converging.
    Returns success or failure based on sense results.

    In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)

  def anomaly(self):
    """
    Called after sense() if the instance has finished converging.
    Returns success or failure based on sense results.
    Failure signals the instance has diverged.

    In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=3, failure_amount=2)



from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise

import datetime
import email.utils
import json
import os
import time
from six.moves.urllib.parse import urlparse
import operator


@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  EXTENDED_STATUS_CODE_MAPPING = {
    '520': 'Too many redirects',
    '523': 'Connection error',
    '524': 'Connection timeout',
    '526': 'SSL Error',

  }

  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # Set frequency compatible to default surykatka interval - 2 minutes
    self.setPeriodicity(float(self.getConfig('frequency', 2)))
    self.failure_amount = int(
      self.getConfig('failure-amount', self.getConfig('failure_amount', 1)))
    self.enabled_sense_list = self.getConfig(
      'enabled-sense-list',
      'dns_query whois tcp_server http_query ssl_certificate'
      ' elapsed_time').split()
    self.result_count = self.failure_amount
    self.error = False
    self.message_list = []
    # Make promise test-less, as it's result is not important for instantiation
    self.setTestLess()

  def appendMessage(self, message):
    self.message_list.append(message)

  def emitLog(self):
   if self.error:
     emit = self.logger.error
   else:
     emit = self.logger.info

   url = self.getConfig('url')
   if url:
     self.message_list.insert(0, '%s :' % (url,))
   emit(' '.join(self.message_list))

  def appendError(self, message):
    self.error = True
    self.message_list.extend(['ERROR', message])

  def appendOk(self, message):
    self.message_list.extend(['OK', message])

  def senseBotStatus(self):
    key = 'bot_status'
    self.appendMessage('%s:' % (key, ))

    if key not in self.surykatka_json:
      self.appendError("%r not in %r" % (key, self.json_file))
      return
    bot_status_list = self.surykatka_json[key]
    if len(bot_status_list) == 0:
      self.appendError("%r empty in %r" % (key, self.json_file))
      return
    bot_status = bot_status_list[0]
    if bot_status.get('text') != 'loop':
      self.appendError(
        "bot_status is %r instead of 'loop' in %r" % (str(
          bot_status.get('text')), self.json_file))
      return
    timetuple = email.utils.parsedate(bot_status['date'])
    last_bot_datetime = datetime.datetime.fromtimestamp(time.mktime(timetuple))
    delta = self.utcnow - last_bot_datetime
    # sanity check
    if delta < datetime.timedelta(minutes=0):
      self.appendError('Last bot datetime is in future')
      return
    if delta > datetime.timedelta(minutes=15):
      self.appendError('Last bot datetime is more than 15 minutes old')
      return

    self.appendOk('Last bot status')

  def senseSslCertificate(self):
    key = 'ssl_certificate'
    self.appendMessage('%s:' % (key, ))

    url = self.getConfig('url')
    parsed_url = urlparse(url)
    if parsed_url.scheme == 'https':
      hostname = parsed_url.netloc
      certificate_expiration_days = self.getConfig(
        'certificate-expiration-days', '15')
      try:
        certificate_expiration_days = int(certificate_expiration_days)
      except ValueError:
        self.appendError(
          'certificate-expiration-days %r is incorrect' % (
            self.getConfig('certificate-expiration-days')))
    else:
      self.appendOk('No check needed')
      return
    if not hostname:
      self.appendError('url is incorrect')
      return
    if key not in self.surykatka_json:
      self.appendError("%r not in %r" % (key, self.json_file))
      return

    entry_list = [
      q for q in self.surykatka_json[key] if q['hostname'] == hostname]
    if len(entry_list) == 0:
      self.appendError('No data')
      return

    for entry in sorted(entry_list, key=operator.itemgetter('ip')):
      timetuple = email.utils.parsedate(entry['not_after'])
      if timetuple is None:
        self.appendError('IP %s no information' % (entry['ip'],))
      else:
        certificate_expiration_time = datetime.datetime.fromtimestamp(
          time.mktime(timetuple))
        if certificate_expiration_time - datetime.timedelta(
          days=certificate_expiration_days) < self.utcnow:
          self.appendError(
            'IP %s expires in < %s days' % (
              entry['ip'], certificate_expiration_days))
        else:
          self.appendOk(
            'IP %s expires in > %s days' % (
              entry['ip'], certificate_expiration_days))

  def senseHttpQuery(self):
    key = 'http_query'
    self.appendMessage('%s:' % (key, ))

    if key not in self.surykatka_json:
      self.appendError("%r not in %r" % (key, self.json_file))
      return

    url = self.getConfig('url')
    status_code = self.getConfig('status-code')
    http_header_dict = json.loads(self.getConfig('http-header-dict', '{}'))

    entry_list = [q for q in self.surykatka_json[key] if q['url'] == url]
    if len(entry_list) == 0:
      self.appendError('No data')
      return

    for entry in sorted(entry_list, key=operator.itemgetter('ip')):
      entry_status_code = str(entry['status_code'])
      if entry_status_code != status_code:
        status_code_explanation = self.EXTENDED_STATUS_CODE_MAPPING.get(
          entry_status_code)
        if status_code_explanation:
          status_code_explanation = '%s (%s)' % (
            entry_status_code, status_code_explanation)
        else:
          status_code_explanation = entry_status_code
        self.appendError(
          'IP %s status_code %s != %s' % (
            entry['ip'], status_code_explanation, status_code))
      else:
        self.appendOk(
          'IP %s status_code %s' % (entry['ip'], status_code))
      if http_header_dict:
        if http_header_dict != entry['http_header_dict']:
          self.appendError(
            'IP %s HTTP Header %s != %s' % (
              entry['ip'],
              json.dumps(http_header_dict, sort_keys=True),
              json.dumps(entry['http_header_dict'], sort_keys=True)))
        else:
          self.appendOk(
            'IP %s HTTP Header %s' % (
              entry['ip'], json.dumps(http_header_dict, sort_keys=True)))

  def senseDnsQuery(self):
    key = 'dns_query'
    self.appendMessage('%s:' % (key, ))

    if key not in self.surykatka_json:
      self.appendError("%r not in %r" % (key, self.json_file))
      return

    url = self.getConfig('url')
    hostname = urlparse(url).hostname
    ip_set = set(self.getConfig('ip-list', '').split())

    entry_dict = {}
    for q in self.surykatka_json[key]:
      if q['domain'] == hostname and q['rdtype'] in ('A', 'AAAA'):
        if not q['resolver_ip'] in entry_dict:
          entry_dict[q['resolver_ip']] = {
            'domain': q['domain'],
            'response_ip_set': set()
          }
        entry_dict[q['resolver_ip']]['response_ip_set'].update([r.strip() for r in q['response'].split(",") if r.strip()])

    if len(ip_set):
      if not entry_dict:
        self.appendError('No data')
        return

      for resolver_ip, entry in sorted(entry_dict.items()):
        response_ip_set = entry['response_ip_set']
        if ip_set != response_ip_set:
          self.appendError(
            "resolver's %s: %s != %s" % (
              resolver_ip, ' '.join(sorted(ip_set)),
              ' '.join(sorted(response_ip_set)) or "empty-reply"))
        else:
          self.appendOk(
            "resolver's %s: %s" % (
              resolver_ip, ' '.join(sorted(ip_set)),))
    else:
      self.appendOk('No check configured')

  def senseTcpServer(self):
    key = 'tcp_server'
    self.appendMessage('%s:' % (key, ))

    if key not in self.surykatka_json:
      self.appendError("%r not in %r" % (key, self.json_file))
      return

    url = self.getConfig('url')
    parsed_url = urlparse(url)
    hostname = parsed_url.hostname
    if parsed_url.port is not None:
      port = parsed_url.port
    else:
      if parsed_url.scheme == 'https':
        port = 443
      else:
        port = 80
    ip_set = set(self.getConfig('ip-list', '').split())
    if len(ip_set) == 0:
      self.appendOk('No check configured')
      return

    entry_list = [
      q for q in self.surykatka_json[key]
      if hostname in [
        r.strip() for r in q['domain'].split(',')] and q['port'] == port]
    if len(entry_list) == 0:
      self.appendError('No data')
      return
    for ip in sorted(ip_set):
      ok = False
      for entry in sorted(entry_list, key=operator.itemgetter('ip')):
        if entry['ip'] == ip:
          if entry['state'] == 'closed':
            ok = False
            break
          if entry['state'] == 'open':
            ok = True
      if ":" in ip:
        ip = "[%s]" % ip
      if ok:
        self.appendOk('IP %s:%s' % (ip, port))
      else:
        self.appendError('IP %s:%s' % (ip, port))

  def senseWhois(self):
    key = 'whois'
    self.appendMessage('%s:' % (key, ))
    url = self.getConfig('url')
    parsed_url = urlparse(url)
    hostname = parsed_url.netloc
    if not hostname:
      self.appendError('url is incorrect')
      return
    domain_expiration_days = self.getConfig(
      'domain-expiration-days', '30')
    try:
      domain_expiration_days = int(domain_expiration_days)
    except ValueError:
      self.appendError(
        'domain-expiration-days %r is incorrect' % (
          self.getConfig('domain-expiration-days')))
      return

    if key not in self.surykatka_json:
      self.appendError("%r not in %r" % (key, self.json_file))
      return

    def checkHostnameDomain(hostname, domain):
      if hostname == domain:
        return True
      elif hostname.endswith('.' + domain):
        return True
      return False

    entry_list = [
      q for q in self.surykatka_json[key]
      if checkHostnameDomain(hostname, q['domain'])]
    if len(entry_list) == 0:
      self.appendError('No data')
      return

    if len(entry_list) > 1:
      self.appendError('Bad data')
      return
    entry = entry_list[0]
    expiration_date = entry['expiration_date']
    if expiration_date is None:
      self.appendError('%s expiration date not avaliable' % (entry['domain'],))
      return
    timetuple = email.utils.parsedate(expiration_date)
    if timetuple is None:
      self.appendError("Can't parse date %s" % (expiration_date,))
    domain_expiration_time = datetime.datetime.fromtimestamp(
      time.mktime(timetuple))
    if domain_expiration_time - datetime.timedelta(
      days=domain_expiration_days) < self.utcnow:
      self.appendError(
        '%s expires in < %s days' % (entry['domain'], domain_expiration_days,))
    else:
      self.appendOk(
        '%s expires in > %s days' % (entry['domain'], domain_expiration_days,))

  def senseElapsedTime(self):
    key = 'elapsed_time'
    self.appendMessage('%s:' % (key, ))
    surykatka_key = 'http_query'

    if surykatka_key not in self.surykatka_json:
      self.appendError("%r not in %r" % (surykatka_key, self.json_file))
      return

    url = self.getConfig('url')
    maximum_elapsed_time = self.getConfig('maximum-elapsed-time')

    entry_list = [
      q for q in self.surykatka_json[surykatka_key] if q['url'] == url]
    if len(entry_list) == 0:
      self.appendError('No data')
      return
    if maximum_elapsed_time:
      found = False
      for entry in sorted(entry_list, key=operator.itemgetter('ip')):
        if 'total_seconds' in entry:
          found = True
          maximum_elapsed_time = float(maximum_elapsed_time)
          if entry['total_seconds'] == 0.:
            self.appendError('IP %s failed to reply' % (entry['ip']))
          elif entry['total_seconds'] > maximum_elapsed_time:
            self.appendError(
              'IP %s replied > %.2fs' %
              (entry['ip'], maximum_elapsed_time))
          else:
            self.appendOk('IP %s replied < %.2fs' % (
              entry['ip'], maximum_elapsed_time))
      if not found:
        self.appendError(
          "No entry with total_seconds found. If the error persist, please "
          "update surykatka")
    else:
      self.appendOk("No check configured")

  def sense(self):
    """
      Sense various information about the given url
    """
    self.utcnow = datetime.datetime.utcnow()

    self.json_file = self.getConfig('json-file', '')
    if not os.path.exists(self.json_file):
      self.appendError('File %r does not exists' % self.json_file)
    else:
      with open(self.json_file) as fh:
        try:
          self.surykatka_json = json.load(fh)
        except Exception:
          self.appendError(
            "loading JSON from %r" % self.json_file)
        else:
          report = self.getConfig('report')
          if report == 'bot_status':
            self.senseBotStatus()
          elif report == 'http_query':
            for check_name, check_method in [
              ('dns_query', self.senseDnsQuery),
              ('whois', self.senseWhois),
              ('tcp_server', self.senseTcpServer),
              ('http_query', self.senseHttpQuery),
              ('ssl_certificate', self.senseSslCertificate),
              ('elapsed_time', self.senseElapsedTime)
            ]:
              if check_name in self.enabled_sense_list:
                check_method()
          else:
            self.appendError(
              "Report %r is not supported" % report)
    self.emitLog()

  def anomaly(self):
    """
      Anomaly returns a TestResult instead of AnomalyResult because we don't
      want to call bang when there is a problem.
      This will need a human intervention.
    """
    return self._test(
      result_count=self.result_count, failure_amount=self.failure_amount)



import re
from .util import JSONPromise, get_json_log_data_interval

from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    self.setPeriodicity(float(self.getConfig('frequency', 1)))
    self.amarisoft_rf_info_log = self.getConfig('amarisoft-rf-info-log')
    self.sdr_devchan = "/dev/sdr%s@%s" % (self.getConfig('sdr_dev'), self.getConfig('sfp_port'))
    self.stats_period = int(self.getConfig('stats-period'))
    self.testing = self.getConfig('testing') == "True"

  def sense(self):
      if self.testing:
          self.logger.info("skipping promise")
          return

      def error(msg): self.logger.error("%s: %s", self.sdr_devchan, msg)
      def info(msg):  self.logger.info ("%s: %s", self.sdr_devchan, msg)

      data_list = get_json_log_data_interval(self.amarisoft_rf_info_log, self.stats_period * 2)
      if len(data_list) < 1:
        error("rf_info: stale data")
        return

      rf_info_text = data_list[0]['rf_info']
      rf_info = self._parse_rf_info(rf_info_text)
      if self.sdr_devchan not in rf_info:
        error("rf_info: no device entry")
        return

      rf_info = rf_info[self.sdr_devchan]
      icpri = rf_info.get('CPRI_option') or rf_info.get('CPRI')
      if icpri is None:
        error("no CPRI feature")
        return

      hw = ("HW" in icpri)
      sw = ("SW" in icpri)
      if not hw:
        error("HW Lock is missing")
      if not sw:
        error("SW Lock is missing")
      if hw and sw:
        info("CPRI locked")

  @staticmethod
  def _parse_rf_info(rf_info_text):  # -> {} /dev/sdrX@Y -> {key: value}
    """_parse_rf_info parses rf_info output into per-SDR-device key->value dictionaries.

       For example:

         TRX SDR driver 2023-09-07, API v15/18
         PCIe CPRI /dev/sdr1@2:
           FPGA vccint: 0.98 V
           FPGA vccaux: 1.77 V
         PCIe CPRI /dev/sdr3@4:
           ABC: 123
           DEF: 4567

       is parsed as {'/dev/sdr1@2': {'FPGA vccint': '0.98 V',  'FPGA vccaux': '1.77 V'},
                     '/dev/sdr3@4': {'ABC': '123',  'DEF': '4567'}}
    """
    rf_info = {}
    cur = None
    for l in rf_info_text.splitlines():
      if not l.startswith(' '):  # possibly start of new /dev entry
        new = True
        if l.startswith('Clock tune:'):
            new = False # 2024-06-15 started to emit 'Clock tune:' without indent
        if new:
            cur = None
            m = re.search(r' (/dev/sdr[^\s]+):\s*$', l)
            if m is None: # not so - ignore the line
              continue

            cur = {}
            sdr_devchan = m.group(1)
            rf_info[sdr_devchan] = cur
            continue

      # indented line - it populates current if it still holds its context
      if cur is None:
        continue

      l = l.lstrip()
      if not l:
          continue  # empty lines are ignore, e.g. empty trailing lines in 2022 format

      if ':' not in l:
          raise ValueError('invalid line %r' % (l,))

      k, v = l.split(':', 1)
      k = k.strip()
      v = v.strip()
      cur[k] = v

    return rf_info

  def test(self):
    """
      Called after sense() if the instance is still converging.
      Returns success or failure based on sense results.

      In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)


  def anomaly(self):
    """
      Called after sense() if the instance has finished converging.
      Returns success or failure based on sense results.
      Failure signals the instance has diverged.

      In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=1, failure_amount=1)



import mock
import os
import time
from datetime import datetime
from datetime import timedelta
from slapos.grid.promise import PromiseError
from slapos.promise.plugin.check_oru_config_log import RunPromise
from . import TestPromisePluginMixin


class TestCheckOruConfigLog(TestPromisePluginMixin):

    promise_name = "check-oru-config-log.py"

    def setUp(self):
        super(TestCheckOruConfigLog, self).setUp()

    def writePromise(self, **kw):
        super(TestCheckOruConfigLog, self).writePromise(self.promise_name,
            "from %s import %s\nextra_config_dict = %r\n"
            % (RunPromise.__module__, RunPromise.__name__, kw))

    def test_promise_success(self):
        self.config_log = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'config.log')
        if os.path.exists(self.config_log):
            os.remove(self.config_log)
        with open(self.config_log, 'w+') as f:
            f.write("""2023-05-23 04:32:48,867 [INFO] Sending edit-config RPC request...
2023-05-23 04:32:49,111 [INFO] Edit-config RPC request sent successfully
"""
            )
        self.writePromise(**{
            'config-log': self.config_log,
        })
        self.configureLauncher()
        self.launcher.run()

    def test_promise_fail(self):
        self.config_log = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'config.log')
        if os.path.exists(self.config_log):
            os.remove(self.config_log)
        with open(self.config_log, 'w') as f:
            f.write("""2023-05-23 04:32:20,110 [INFO] Connecting to ('2a11:9ac1:6:800a::1', 830), user oranuser...
2023-05-23 04:32:48,863 [INFO] Connection to ('2a11:9ac1:6:800a::1', 830) successful
2023-05-23 04:32:48,867 [INFO] Sending edit-config RPC request...
2023-05-23 04:32:49,111 [ERROR] Error sending edit-config RPC request: Operation failed
"""
            )
        self.writePromise(**{
            'config-log': self.config_log,
        })
        self.configureLauncher()
        with self.assertRaises(PromiseError):
            self.launcher.run()


if __name__ == '__main__':
    unittest.main()


import subprocess

from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def __init__(self, config):

    super(RunPromise, self).__init__(config)
    self.setPeriodicity(float(self.getConfig('frequency', 1)))

  def sense(self):

    testing = self.getConfig('testing') == "True"
    sdr = self.getConfig('sdr')
    sdr_dev = self.getConfig('sdr_dev')
    dma_chan = self.getConfig('dma_chan')
    sdr_devchan = "/dev/sdr%s@%s" % (sdr_dev, dma_chan)

    if testing:
        self.logger.info("skipping promise")
        return
    try:
      out = subprocess.check_output([
        sdr + '/sdr_util', '-c', sdr_dev, '-d', dma_chan, 'version'], stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
      if e.returncode == 1 and \
        ("DMA channel is already opened" in e.output.decode() or \
         "Device or resource busy" in e.output.decode()):
        self.logger.info("eNB is using %s", sdr_devchan)
        return
    self.logger.error("eNB is not using %s", sdr_devchan)

  def test(self):
    """
      Called after sense() if the instance is still converging.
      Returns success or failure based on sense results.

      In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)

  def anomaly(self):
    """
      Called after sense() if the instance has finished converging.
      Returns success or failure based on sense results.
      Failure signals the instance has diverged.

      In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=1, failure_amount=1)



import errno
import json
import logging
import os

from dateutil import parser
from .util import JSONPromise
from .util import tail_file

from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(JSONPromise):
    def __init__(self, config):
        super(RunPromise, self).__init__(config)
        self.setPeriodicity(float(self.getConfig('frequency', 1)))
        self.stats_log = self.getConfig('stats-log')
        self.testing = self.getConfig('testing') == "True"

    def sense(self):

        if self.testing:
            self.logger.info("skipping promise")
            return

        latest_log = tail_file(self.stats_log)
        latest_log.split("\n")

        if "Waiting for notification from" not in latest_log:
            self.logger.error("Not subscribed")
        else:
            self.logger.info("Subscribed successful")
          
    def test(self):
        """
        Called after sense() if the instance is still converging.
        Returns success or failure based on sense results.

        In this case, fail if the previous sensor result is negative.
        """
        return self._test(result_count=1, failure_amount=1)

    def anomaly(self):
        """
        Called after sense() if the instance has finished converging.
        Returns success or failure based on sense results.
        Failure signals the instance has diverged.

        In this case, fail if two out of the last three results are negative.
        """
        return self._anomaly(result_count=1, failure_amount=1)



from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def __init__(self, config):
    """
      Called when initialising the promise before testing.
      Sets the configuration and the periodicity.
    """
    super(RunPromise, self).__init__(config)
    self.setPeriodicity(float(self.getConfig('frequency', 1)))


  def sense(self):
    """
      Called every time the promise is tested.
      Signals a positive or negative result.

      In this case, check whether the file exists.
    """
    frequency = self.getConfig('frequency')
    range_rating = self.getConfig('range-rating')
    try:
      min_frequency = int(range_rating.split('MHz')[0].strip())
      max_frequency = int(range_rating.split('-')[1].split('MHz')[0].strip())
    except (IndexError, ValueError) as e:
      self.logger.info("Range rating not available, skipping the promise")
      return
    try:
      frequency = int(float(frequency))
    except ValueError as e:
      self.logger.info("Invalid frequency, skipping the promise")
      return

    if min_frequency <= frequency <= max_frequency:
      self.logger.info("Frequency is in bounds ({} MHz <= {} MHz <= {} MHz)".format(
        min_frequency,
        frequency,
        max_frequency))
    elif frequency < min_frequency:
      self.logger.error("Frequency is lower than the lowest possible frequency on this hardware, please increase it ({} MHz < {} MHz)".format(
        frequency,
        min_frequency))
    else:
      self.logger.error("Frequency is higher than the highest possible frequency on this hardware, please increase it ({} MHz > {} MHz)".format(
        frequency,
        max_frequency))

  def test(self):
    """
      Called after sense() if the instance is still converging.
      Returns success or failure based on sense results.

      In this case, fail if the previous sensor result is negative.
    """
    return self._test(result_count=1, failure_amount=1)


  def anomaly(self):
    """
      Called after sense() if the instance has finished converging.
      Returns success or failure based on sense results.
      Failure signals the instance has diverged.

      In this case, fail if two out of the last three results are negative.
    """
    return self._anomaly(result_count=1, failure_amount=1)



# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2022 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
##############################################################################
import mock

from collections import namedtuple
from slapos.grid.promise import PromiseError
from slapos.promise.plugin.check_network_errors_packets import RunPromise
from . import TestPromisePluginMixin


class TestCheckNetwork(TestPromisePluginMixin):

  promise_name = "monitor-network.py"

  def setUp(self):
    super(TestCheckNetwork, self).setUp()
    self.network_data = namedtuple('network_data', 
    ['bytes_recv', 'bytes_sent','errin', 'errout', 'dropin', 'dropout'])

  def writePromise(self, **kw):
    super(TestCheckNetwork, self).writePromise(self.promise_name,
      "from %s import %s\nextra_config_dict = %r\n"
      % (RunPromise.__module__, RunPromise.__name__, kw))

  def runPromise(self, summary, failed=False):
    self.configureLauncher(enable_anomaly=True, force=True)
    with mock.patch('psutil.net_io_counters', return_value=summary):
      if failed:
        self.assertRaises(PromiseError, self.launcher.run)
      else:
        self.launcher.run()
    result = self.getPromiseResult(self.promise_name)['result']
    self.assertEqual(result['failed'], failed)
    return result['message']

  def test_network_ok(self):
    message = "Network statistics OK"
    mock_stats = {'bytes_recv':0, 'bytes_sent':0,'errin':0, 'errout':0, 'dropin':0, 'dropout':0}
    self.writePromise(**{
        'max-lost-packets-per-MB': 5,
        'max-error-messages-per-MB': 5,
    })
    self.assertEqual(message, self.runPromise(self.network_data(**mock_stats)))

  def test_network_dropped_packets_nok(self):
    message = "Network packets lost reached critical threshold: 10  (threshold is 5 per MB)"
    mock_stats = {'bytes_recv':0, 'bytes_sent':0,'errin':0, 'errout':0, 'dropin':5, 'dropout':5}
    self.writePromise(**{
        'max-lost-packets-per-MB': 5,
        'max-error-messages-per-MB': 5,
    })
    self.assertEqual(message, self.runPromise(self.network_data(**mock_stats)))

  def test_network_errors_nok(self):
    message = "Network errors reached critical threshold: 10  (threshold is 5 per MB)"
    mock_stats = {'bytes_recv':0, 'bytes_sent':0,'errin':5, 'errout':5, 'dropin':0, 'dropout':0}
    self.writePromise(**{
        'max-lost-packets-per-MB': 5,
        'max-error-messages-per-MB': 5,
    })
    self.assertEqual(message, self.runPromise(self.network_data(**mock_stats)))

  def test_network_nok(self):
    message = "Network packets lost reached critical threshold: 10  (threshold is 5 per MB)"\
      "\nNetwork errors reached critical threshold: 10  (threshold is 5 per MB)"
    mock_stats = {'bytes_recv':0, 'bytes_sent':0, 'errin':5, 'errout':5, 'dropin':5, 'dropout':5}
    self.writePromise(**{
        'max-lost-packets-per-MB': 5,
        'max-error-messages-per-MB': 5,
    })
    self.assertEqual(message, self.runPromise(self.network_data(**mock_stats)))

if __name__ == '__main__':
  unittest.main()



from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise.generic import GenericPromise
import os


@implementer(interface.IPromise)
class RunPromise(GenericPromise):
  def __init__(self, config):
    super(RunPromise, self).__init__(config)
    # SR can set custom periodicity
    self.setPeriodicity(float(self.getConfig('frequency', 2)))
    self.result_count = int(self.getConfig('result-count', '1'))
    self.failure_amount = int(self.getConfig('failure-amount', '1'))
    if self.result_count < self.failure_amount:
      raise ValueError(
        'Bad configuration: result-count %i < failure_amount %i' % (
          self.result_count, self.failure_amount))

    if self.getConfig(
      'perdiodic-only', 'false').lower() in ('true', 'yes', '1'):
      self.setTestLess()

    if self.getConfig(
      'report-anomaly', 'true').lower() in ('false', 'no', '0'):
      self.setAnomalyLess()

  def sense(self):
    """
      Check state of the filename

      state can be one of:
        absent
        empty
        not-empty
    """

    filename = self.getConfig('filename')
    state = self.getConfig('state')
    url = (self.getConfig('url') or '').strip()

    exists = os.path.exists(filename)
    if state == 'absent':
      if exists:
        self.logger.error("ERROR %r not absent", filename)
      else:
        self.logger.info("OK %r state %r" % (filename, state))
      return
    if not exists:
      self.logger.error("ERROR %r not present", filename)
      return

    try:
      with open(filename) as f:
        result = f.read()
    except Exception as e:
      self.logger.error(
        "ERROR %r during opening and reading file %r" % (e, filename))
      return

    if state == 'empty' and result != '':
      message_list = ['ERROR %r not empty' % (filename,)]
      if url:
        message_list.append(', content available at %s' % (url,))
      self.logger.error(''.join(message_list))
    elif state == 'not-empty' and result == '':
      self.logger.error(
          "ERROR %r empty" % (filename,))
    else:
      self.logger.info("OK %r state %r" % (filename, state))

  def test(self):
    return self._test(
      result_count=self.result_count, failure_amount=self.failure_amount)

  def anomaly(self):
    return self._anomaly(
      result_count=self.result_count, failure_amount=self.failure_amount)
